//@HEADER
// ************************************************************************
//
//                        Kokkos v. 4.0
//       Copyright (2022) National Technology & Engineering
//               Solutions of Sandia, LLC (NTESS).
//
// Under the terms of Contract DE-NA0003525 with NTESS,
// the U.S. Government retains certain rights in this software.
//
// Part of Kokkos, under the Apache License v2.0 with LLVM Exceptions.
// See https://kokkos.org/LICENSE for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//@HEADER

#ifndef KOKKOSSPARSE_SPMV_BSRMATRIX_TPL_SPEC_DECL_HPP
#define KOKKOSSPARSE_SPMV_BSRMATRIX_TPL_SPEC_DECL_HPP

#include "KokkosKernels_Controls.hpp"
#include "KokkosSparse_Utils_mkl.hpp"

#ifdef KOKKOSKERNELS_ENABLE_TPL_MKL
#include <mkl.h>

namespace KokkosSparse {
namespace Experimental {
namespace Impl {

#if (__INTEL_MKL__ > 2017)
// MKL 2018 and above: use new interface: sparse_matrix_t and mkl_sparse_?_mv()

using KokkosSparse::Impl::mode_kk_to_mkl;

inline matrix_descr getDescription() {
  matrix_descr A_descr;
  A_descr.type = SPARSE_MATRIX_TYPE_GENERAL;
  A_descr.mode = SPARSE_FILL_MODE_FULL;
  A_descr.diag = SPARSE_DIAG_NON_UNIT;
  return A_descr;
}

inline void spmv_block_impl_mkl(sparse_operation_t op, float alpha, float beta,
                                int m, int n, int b, const int* Arowptrs,
                                const int* Aentries, const float* Avalues,
                                const float* x, float* y) {
  sparse_matrix_t A_mkl;
  KOKKOSKERNELS_MKL_SAFE_CALL(mkl_sparse_s_create_bsr(
      &A_mkl, SPARSE_INDEX_BASE_ZERO, SPARSE_LAYOUT_ROW_MAJOR, m, n, b,
      const_cast<int*>(Arowptrs), const_cast<int*>(Arowptrs + 1),
      const_cast<int*>(Aentries), const_cast<float*>(Avalues)));

  matrix_descr A_descr = getDescription();
  KOKKOSKERNELS_MKL_SAFE_CALL(
      mkl_sparse_s_mv(op, alpha, A_mkl, A_descr, x, beta, y));
}

inline void spmv_block_impl_mkl(sparse_operation_t op, double alpha,
                                double beta, int m, int n, int b,
                                const int* Arowptrs, const int* Aentries,
                                const double* Avalues, const double* x,
                                double* y) {
  sparse_matrix_t A_mkl;
  KOKKOSKERNELS_MKL_SAFE_CALL(mkl_sparse_d_create_bsr(
      &A_mkl, SPARSE_INDEX_BASE_ZERO, SPARSE_LAYOUT_ROW_MAJOR, m, n, b,
      const_cast<int*>(Arowptrs), const_cast<int*>(Arowptrs + 1),
      const_cast<int*>(Aentries), const_cast<double*>(Avalues)));

  matrix_descr A_descr = getDescription();
  KOKKOSKERNELS_MKL_SAFE_CALL(
      mkl_sparse_d_mv(op, alpha, A_mkl, A_descr, x, beta, y));
}

inline void spmv_block_impl_mkl(sparse_operation_t op,
                                Kokkos::complex<float> alpha,
                                Kokkos::complex<float> beta, int m, int n,
                                int b, const int* Arowptrs, const int* Aentries,
                                const Kokkos::complex<float>* Avalues,
                                const Kokkos::complex<float>* x,
                                Kokkos::complex<float>* y) {
  sparse_matrix_t A_mkl;
  KOKKOSKERNELS_MKL_SAFE_CALL(mkl_sparse_c_create_bsr(
      &A_mkl, SPARSE_INDEX_BASE_ZERO, SPARSE_LAYOUT_ROW_MAJOR, m, n, b,
      const_cast<int*>(Arowptrs), const_cast<int*>(Arowptrs + 1),
      const_cast<int*>(Aentries), (MKL_Complex8*)Avalues));

  MKL_Complex8 alpha_mkl{alpha.real(), alpha.imag()};
  MKL_Complex8 beta_mkl{beta.real(), beta.imag()};
  matrix_descr A_descr = getDescription();
  KOKKOSKERNELS_MKL_SAFE_CALL(mkl_sparse_c_mv(
      op, alpha_mkl, A_mkl, A_descr, reinterpret_cast<const MKL_Complex8*>(x),
      beta_mkl, reinterpret_cast<MKL_Complex8*>(y)));
}

inline void spmv_block_impl_mkl(sparse_operation_t op,
                                Kokkos::complex<double> alpha,
                                Kokkos::complex<double> beta, int m, int n,
                                int b, const int* Arowptrs, const int* Aentries,
                                const Kokkos::complex<double>* Avalues,
                                const Kokkos::complex<double>* x,
                                Kokkos::complex<double>* y) {
  sparse_matrix_t A_mkl;
  KOKKOSKERNELS_MKL_SAFE_CALL(mkl_sparse_z_create_bsr(
      &A_mkl, SPARSE_INDEX_BASE_ZERO, SPARSE_LAYOUT_ROW_MAJOR, m, n, b,
      const_cast<int*>(Arowptrs), const_cast<int*>(Arowptrs + 1),
      const_cast<int*>(Aentries), (MKL_Complex16*)Avalues));

  matrix_descr A_descr = getDescription();
  MKL_Complex16 alpha_mkl{alpha.real(), alpha.imag()};
  MKL_Complex16 beta_mkl{beta.real(), beta.imag()};
  KOKKOSKERNELS_MKL_SAFE_CALL(mkl_sparse_z_mv(
      op, alpha_mkl, A_mkl, A_descr, reinterpret_cast<const MKL_Complex16*>(x),
      beta_mkl, reinterpret_cast<MKL_Complex16*>(y)));
}

inline void spm_mv_block_impl_mkl(sparse_operation_t op, float alpha,
                                  float beta, int m, int n, int b,
                                  const int* Arowptrs, const int* Aentries,
                                  const float* Avalues, const float* x,
                                  int colx, int ldx, float* y, int ldy) {
  sparse_matrix_t A_mkl;
  KOKKOSKERNELS_MKL_SAFE_CALL(mkl_sparse_s_create_bsr(
      &A_mkl, SPARSE_INDEX_BASE_ZERO, SPARSE_LAYOUT_ROW_MAJOR, m, n, b,
      const_cast<int*>(Arowptrs), const_cast<int*>(Arowptrs + 1),
      const_cast<int*>(Aentries), const_cast<float*>(Avalues)));

  matrix_descr A_descr = getDescription();
  KOKKOSKERNELS_MKL_SAFE_CALL(mkl_sparse_s_mm(op, alpha, A_mkl, A_descr,
                                              SPARSE_LAYOUT_ROW_MAJOR, x, colx,
                                              ldx, beta, y, ldy));
}

inline void spm_mv_block_impl_mkl(sparse_operation_t op, double alpha,
                                  double beta, int m, int n, int b,
                                  const int* Arowptrs, const int* Aentries,
                                  const double* Avalues, const double* x,
                                  int colx, int ldx, double* y, int ldy) {
  sparse_matrix_t A_mkl;
  KOKKOSKERNELS_MKL_SAFE_CALL(mkl_sparse_d_create_bsr(
      &A_mkl, SPARSE_INDEX_BASE_ZERO, SPARSE_LAYOUT_ROW_MAJOR, m, n, b,
      const_cast<int*>(Arowptrs), const_cast<int*>(Arowptrs + 1),
      const_cast<int*>(Aentries), const_cast<double*>(Avalues)));

  matrix_descr A_descr = getDescription();
  KOKKOSKERNELS_MKL_SAFE_CALL(mkl_sparse_d_mm(op, alpha, A_mkl, A_descr,
                                              SPARSE_LAYOUT_ROW_MAJOR, x, colx,
                                              ldx, beta, y, ldy));
}

inline void spm_mv_block_impl_mkl(sparse_operation_t op,
                                  Kokkos::complex<float> alpha,
                                  Kokkos::complex<float> beta, int m, int n,
                                  int b, const int* Arowptrs,
                                  const int* Aentries,
                                  const Kokkos::complex<float>* Avalues,
                                  const Kokkos::complex<float>* x, int colx,
                                  int ldx, Kokkos::complex<float>* y, int ldy) {
  sparse_matrix_t A_mkl;
  KOKKOSKERNELS_MKL_SAFE_CALL(mkl_sparse_c_create_bsr(
      &A_mkl, SPARSE_INDEX_BASE_ZERO, SPARSE_LAYOUT_ROW_MAJOR, m, n, b,
      const_cast<int*>(Arowptrs), const_cast<int*>(Arowptrs + 1),
      const_cast<int*>(Aentries), (MKL_Complex8*)Avalues));

  MKL_Complex8 alpha_mkl{alpha.real(), alpha.imag()};
  MKL_Complex8 beta_mkl{beta.real(), beta.imag()};
  matrix_descr A_descr = getDescription();
  KOKKOSKERNELS_MKL_SAFE_CALL(
      mkl_sparse_c_mm(op, alpha_mkl, A_mkl, A_descr, SPARSE_LAYOUT_ROW_MAJOR,
                      reinterpret_cast<const MKL_Complex8*>(x), colx, ldx,
                      beta_mkl, reinterpret_cast<MKL_Complex8*>(y), ldy));
}

inline void spm_mv_block_impl_mkl(
    sparse_operation_t op, Kokkos::complex<double> alpha,
    Kokkos::complex<double> beta, int m, int n, int b, const int* Arowptrs,
    const int* Aentries, const Kokkos::complex<double>* Avalues,
    const Kokkos::complex<double>* x, int colx, int ldx,
    Kokkos::complex<double>* y, int ldy) {
  sparse_matrix_t A_mkl;
  KOKKOSKERNELS_MKL_SAFE_CALL(mkl_sparse_z_create_bsr(
      &A_mkl, SPARSE_INDEX_BASE_ZERO, SPARSE_LAYOUT_ROW_MAJOR, m, n, b,
      const_cast<int*>(Arowptrs), const_cast<int*>(Arowptrs + 1),
      const_cast<int*>(Aentries), (MKL_Complex16*)Avalues));

  matrix_descr A_descr = getDescription();
  MKL_Complex16 alpha_mkl{alpha.real(), alpha.imag()};
  MKL_Complex16 beta_mkl{beta.real(), beta.imag()};
  KOKKOSKERNELS_MKL_SAFE_CALL(
      mkl_sparse_z_mm(op, alpha_mkl, A_mkl, A_descr, SPARSE_LAYOUT_ROW_MAJOR,
                      reinterpret_cast<const MKL_Complex16*>(x), colx, ldx,
                      beta_mkl, reinterpret_cast<MKL_Complex16*>(y), ldy));
}

#endif

#if (__INTEL_MKL__ == 2017)

inline void spmv_block_impl_mkl(char mode, float alpha, float beta, int m,
                                int n, int b, const int* Arowptrs,
                                const int* Aentries, const float* Avalues,
                                const float* x, float* y) {
  mkl_sbsrmv(&mode, &m, &n, &b, &alpha, "G**C", Avalues, Aentries, Arowptrs,
             Arowptrs + 1, x, &beta, y);
}

inline void spmv_block_impl_mkl(char mode, double alpha, double beta, int m,
                                int n, int b, const int* Arowptrs,
                                const int* Aentries, const double* Avalues,
                                const double* x, double* y) {
  mkl_dbsrmv(&mode, &m, &n, &b, &alpha, "G**C", Avalues, Aentries, Arowptrs,
             Arowptrs + 1, x, &beta, y);
}

inline void spmv_block_impl_mkl(char mode, Kokkos::complex<float> alpha,
                                Kokkos::complex<float> beta, int m, int n,
                                int b, const int* Arowptrs, const int* Aentries,
                                const Kokkos::complex<float>* Avalues,
                                const Kokkos::complex<float>* x,
                                Kokkos::complex<float>* y) {
  const MKL_Complex8* alpha_mkl = reinterpret_cast<const MKL_Complex8*>(&alpha);
  const MKL_Complex8* beta_mkl  = reinterpret_cast<const MKL_Complex8*>(&beta);
  const MKL_Complex8* Avalues_mkl =
      reinterpret_cast<const MKL_Complex8*>(Avalues);
  const MKL_Complex8* x_mkl = reinterpret_cast<const MKL_Complex8*>(x);
  MKL_Complex8* y_mkl       = reinterpret_cast<MKL_Complex8*>(y);
  mkl_cbsrmv(&mode, &m, &n, &b, alpha_mkl, "G**C", Avalues_mkl, Aentries,
             Arowptrs, Arowptrs + 1, x_mkl, beta_mkl, y_mkl);
}

inline void spmv_block_impl_mkl(char mode, Kokkos::complex<double> alpha,
                                Kokkos::complex<double> beta, int m, int n,
                                int b, const int* Arowptrs, const int* Aentries,
                                const Kokkos::complex<double>* Avalues,
                                const Kokkos::complex<double>* x,
                                Kokkos::complex<double>* y) {
  const MKL_Complex16* alpha_mkl =
      reinterpret_cast<const MKL_Complex16*>(&alpha);
  const MKL_Complex16* beta_mkl = reinterpret_cast<const MKL_Complex16*>(&beta);
  const MKL_Complex16* Avalues_mkl =
      reinterpret_cast<const MKL_Complex16*>(Avalues);
  const MKL_Complex16* x_mkl = reinterpret_cast<const MKL_Complex16*>(x);
  MKL_Complex16* y_mkl       = reinterpret_cast<MKL_Complex16*>(y);
  mkl_zbsrmv(&mode, &m, &n, &b, alpha_mkl, "G**C", Avalues_mkl, Aentries,
             Arowptrs, Arowptrs + 1, x_mkl, beta_mkl, y_mkl);
}

inline void spm_mv_block_impl_mkl(char mode, float alpha, float beta, int m,
                                  int n, int b, const int* Arowptrs,
                                  const int* Aentries, const float* Avalues,
                                  const float* x, int colx, int ldx, float* y,
                                  int ldy) {
  mkl_sbsrmm(&mode, &m, &n, &colx, &b, &alpha, "G**C", Avalues, Aentries,
             Arowptrs, Arowptrs + 1, x, &beta, y);
}

inline void spm_mv_block_impl_mkl(char mode, double alpha, double beta, int m,
                                  int n, int b, const int* Arowptrs,
                                  const int* Aentries, const double* Avalues,
                                  const double* x, int colx, int ldx, double* y,
                                  int ldy) {
  mkl_dbsrmm(&mode, &m, &n, &colx, &b, &alpha, "G**C", Avalues, Aentries,
             Arowptrs, Arowptrs + 1, x, ldx, &beta, y, ldy);
}

inline void spm_mv_block_impl_mkl(char mode, Kokkos::complex<float> alpha,
                                  Kokkos::complex<float> beta, int m, int n,
                                  int b, const int* Arowptrs,
                                  const int* Aentries,
                                  const Kokkos::complex<float>* Avalues,
                                  const Kokkos::complex<float>* x, int colx,
                                  int ldx, Kokkos::complex<float>* y, int ldy) {
  const MKL_Complex8* alpha_mkl = reinterpret_cast<const MKL_Complex8*>(&alpha);
  const MKL_Complex8* beta_mkl  = reinterpret_cast<const MKL_Complex8*>(&beta);
  const MKL_Complex8* Avalues_mkl =
      reinterpret_cast<const MKL_Complex8*>(Avalues);
  const MKL_Complex8* x_mkl = reinterpret_cast<const MKL_Complex8*>(x);
  MKL_Complex8* y_mkl       = reinterpret_cast<MKL_Complex8*>(y);
  mkl_cbsrmv(&mode, &m, &n, &colx, &b, alpha_mkl, "G**C", Avalues_mkl, Aentries,
             Arowptrs, Arowptrs + 1, x_mkl, ldx, beta_mkl, y_mkl, ldy);
}

inline void spm_mv_block_impl_mkl(
    char mode, Kokkos::complex<double> alpha, Kokkos::complex<double> beta,
    int m, int n, int b, const int* Arowptrs, const int* Aentries,
    const Kokkos::complex<double>* Avalues, const Kokkos::complex<double>* x,
    int colx, int ldx, Kokkos::complex<double>* y, int ldy) {
  const MKL_Complex16* alpha_mkl =
      reinterpret_cast<const MKL_Complex16*>(&alpha);
  const MKL_Complex16* beta_mkl = reinterpret_cast<const MKL_Complex16*>(&beta);
  const MKL_Complex16* Avalues_mkl =
      reinterpret_cast<const MKL_Complex16*>(Avalues);
  const MKL_Complex16* x_mkl = reinterpret_cast<const MKL_Complex16*>(x);
  MKL_Complex16* y_mkl       = reinterpret_cast<MKL_Complex16*>(y);
  mkl_zbsrmv(&mode, &m, &n, &colx, &b, alpha_mkl, "G**C", Avalues_mkl, Aentries,
             Arowptrs, Arowptrs + 1, x_mkl, ldx, beta_mkl, y_mkl, ldy);
}

#endif

#define KOKKOSSPARSE_SPMV_MKL(SCALAR, EXECSPACE, COMPILE_LIBRARY)              \
  template <>                                                                  \
  struct SPMV_BSRMATRIX<                                                       \
      SCALAR const, int const, Kokkos::Device<EXECSPACE, Kokkos::HostSpace>,   \
      Kokkos::MemoryTraits<Kokkos::Unmanaged>, int const, SCALAR const*,       \
      Kokkos::LayoutLeft, Kokkos::Device<EXECSPACE, Kokkos::HostSpace>,        \
      Kokkos::MemoryTraits<Kokkos::Unmanaged | Kokkos::RandomAccess>, SCALAR*, \
      Kokkos::LayoutLeft, Kokkos::Device<EXECSPACE, Kokkos::HostSpace>,        \
      Kokkos::MemoryTraits<Kokkos::Unmanaged>, true, COMPILE_LIBRARY> {        \
    using device_type = Kokkos::Device<EXECSPACE, Kokkos::HostSpace>;          \
    using AMatrix =                                                            \
        BsrMatrix<SCALAR const, int const, device_type,                        \
                  Kokkos::MemoryTraits<Kokkos::Unmanaged>, int const>;         \
    using XVector = Kokkos::View<                                              \
        SCALAR const*, Kokkos::LayoutLeft, device_type,                        \
        Kokkos::MemoryTraits<Kokkos::Unmanaged | Kokkos::RandomAccess>>;       \
    using YVector = Kokkos::View<SCALAR*, Kokkos::LayoutLeft, device_type,     \
                                 Kokkos::MemoryTraits<Kokkos::Unmanaged>>;     \
    using coefficient_type = typename YVector::non_const_value_type;           \
                                                                               \
    static void spmv_bsrmatrix(                                                \
        const KokkosKernels::Experimental::Controls& /*controls*/,             \
        const char mode[], const coefficient_type& alpha, const AMatrix& A,    \
        const XVector& X, const coefficient_type& beta, const YVector& Y) {    \
      std::string label = "KokkosSparse::spmv[TPL_MKL,BSRMATRIX" +             \
                          Kokkos::ArithTraits<SCALAR>::name() + "]";           \
      Kokkos::Profiling::pushRegion(label);                                    \
      spmv_block_impl_mkl(mode_kk_to_mkl(mode[0]), alpha, beta, A.numRows(),   \
                          A.numCols(), A.blockDim(), A.graph.row_map.data(),   \
                          A.graph.entries.data(), A.values.data(), X.data(),   \
                          Y.data());                                           \
      Kokkos::Profiling::popRegion();                                          \
    }                                                                          \
  };

#ifdef KOKKOS_ENABLE_SERIAL
KOKKOSSPARSE_SPMV_MKL(float, Kokkos::Serial, KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_MKL(double, Kokkos::Serial,
                      KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_MKL(Kokkos::complex<float>, Kokkos::Serial,
                      KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_MKL(Kokkos::complex<double>, Kokkos::Serial,
                      KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
#endif

#ifdef KOKKOS_ENABLE_OPENMP
KOKKOSSPARSE_SPMV_MKL(float, Kokkos::OpenMP, KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_MKL(double, Kokkos::OpenMP,
                      KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_MKL(Kokkos::complex<float>, Kokkos::OpenMP,
                      KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_MKL(Kokkos::complex<double>, Kokkos::OpenMP,
                      KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
#endif

#undef KOKKOSSPARSE_SPMV_MKL

#define KOKKOSSPARSE_SPMV_MV_MKL(SCALAR, EXECSPACE, COMPILE_LIBRARY)           \
  template <>                                                                  \
  struct SPMV_MV_BSRMATRIX<                                                    \
      SCALAR const, int const, Kokkos::Device<EXECSPACE, Kokkos::HostSpace>,   \
      Kokkos::MemoryTraits<Kokkos::Unmanaged>, int const, SCALAR const**,      \
      Kokkos::LayoutLeft, Kokkos::Device<EXECSPACE, Kokkos::HostSpace>,        \
      Kokkos::MemoryTraits<Kokkos::Unmanaged | Kokkos::RandomAccess>,          \
      SCALAR**, Kokkos::LayoutLeft,                                            \
      Kokkos::Device<EXECSPACE, Kokkos::HostSpace>,                            \
      Kokkos::MemoryTraits<Kokkos::Unmanaged>, true, true, COMPILE_LIBRARY> {  \
    using device_type = Kokkos::Device<EXECSPACE, Kokkos::HostSpace>;          \
    using AMatrix =                                                            \
        BsrMatrix<SCALAR const, int const, device_type,                        \
                  Kokkos::MemoryTraits<Kokkos::Unmanaged>, int const>;         \
    using XVector = Kokkos::View<                                              \
        SCALAR const**, Kokkos::LayoutLeft, device_type,                       \
        Kokkos::MemoryTraits<Kokkos::Unmanaged | Kokkos::RandomAccess>>;       \
    using YVector = Kokkos::View<SCALAR**, Kokkos::LayoutLeft, device_type,    \
                                 Kokkos::MemoryTraits<Kokkos::Unmanaged>>;     \
    using coefficient_type = typename YVector::non_const_value_type;           \
                                                                               \
    static void spmv_mv_bsrmatrix(                                             \
        const KokkosKernels::Experimental::Controls& /*controls*/,             \
        const char mode[], const coefficient_type& alpha, const AMatrix& A,    \
        const XVector& X, const coefficient_type& beta, const YVector& Y) {    \
      std::string label = "KokkosSparse::spmv[TPL_MKL,BSRMATRIX" +             \
                          Kokkos::ArithTraits<SCALAR>::name() + "]";           \
      Kokkos::Profiling::pushRegion(label);                                    \
      int colx = static_cast<int>(X.extent(1));                                \
      int ldx  = static_cast<int>(X.stride_1());                               \
      int ldy  = static_cast<int>(Y.stride_1());                               \
      spm_mv_block_impl_mkl(mode_kk_to_mkl(mode[0]), alpha, beta, A.numRows(), \
                            A.numCols(), A.blockDim(), A.graph.row_map.data(), \
                            A.graph.entries.data(), A.values.data(), X.data(), \
                            colx, ldx, Y.data(), ldy);                         \
      Kokkos::Profiling::popRegion();                                          \
    }                                                                          \
  };

#ifdef KOKKOS_ENABLE_SERIAL
KOKKOSSPARSE_SPMV_MV_MKL(float, Kokkos::Serial,
                         KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_MV_MKL(double, Kokkos::Serial,
                         KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_MV_MKL(Kokkos::complex<float>, Kokkos::Serial,
                         KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_MV_MKL(Kokkos::complex<double>, Kokkos::Serial,
                         KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
#endif

#ifdef KOKKOS_ENABLE_OPENMP
KOKKOSSPARSE_SPMV_MV_MKL(float, Kokkos::OpenMP,
                         KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_MV_MKL(double, Kokkos::OpenMP,
                         KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_MV_MKL(Kokkos::complex<float>, Kokkos::OpenMP,
                         KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_MV_MKL(Kokkos::complex<double>, Kokkos::OpenMP,
                         KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
#endif

#undef KOKKOSSPARSE_SPMV_MV_MKL

}  // namespace Impl
}  // namespace Experimental
}  // namespace KokkosSparse

#endif

// cuSPARSE
#ifdef KOKKOSKERNELS_ENABLE_TPL_CUSPARSE
#include "cusparse.h"
#include "KokkosSparse_Utils_cusparse.hpp"

//
// From  https://docs.nvidia.com/cuda/cusparse/index.html#bsrmv
// Several comments on bsrmv():
// - Only blockDim > 1 is supported
// - Only CUSPARSE_OPERATION_NON_TRANSPOSE is supported
// - Only CUSPARSE_MATRIX_TYPE_GENERAL is supported.
//
namespace KokkosSparse {
namespace Experimental {
namespace Impl {

template <class AMatrix, class XVector, class YVector>
void spmv_block_impl_cusparse(
    const KokkosKernels::Experimental::Controls& controls, const char mode[],
    typename YVector::non_const_value_type const& alpha, const AMatrix& A,
    const XVector& x, typename YVector::non_const_value_type const& beta,
    const YVector& y) {
  using offset_type = typename AMatrix::non_const_size_type;
  using entry_type  = typename AMatrix::non_const_ordinal_type;
  using value_type  = typename AMatrix::non_const_value_type;

  /* initialize cusparse library */
  cusparseHandle_t cusparseHandle = controls.getCusparseHandle();

  /* Set the operation mode */
  cusparseOperation_t myCusparseOperation;
  switch (toupper(mode[0])) {
    case 'N': myCusparseOperation = CUSPARSE_OPERATION_NON_TRANSPOSE; break;
    default: {
      std::cerr << "Mode " << mode << " invalid for cusparse[*]bsrmv.\n";
      throw std::invalid_argument("Invalid mode");
    }
  }

#if (9000 <= CUDA_VERSION)

  /* create and set the matrix descriptor */
  cusparseMatDescr_t descrA = 0;
  KOKKOS_CUSPARSE_SAFE_CALL(cusparseCreateMatDescr(&descrA));
  KOKKOS_CUSPARSE_SAFE_CALL(
      cusparseSetMatType(descrA, CUSPARSE_MATRIX_TYPE_GENERAL));
  KOKKOS_CUSPARSE_SAFE_CALL(
      cusparseSetMatIndexBase(descrA, CUSPARSE_INDEX_BASE_ZERO));
  cusparseDirection_t dirA = CUSPARSE_DIRECTION_ROW;

  /* perform the actual SpMV operation */
  if ((std::is_same<int, offset_type>::value) &&
      (std::is_same<int, entry_type>::value)) {
    if (std::is_same<value_type, float>::value) {
      KOKKOS_CUSPARSE_SAFE_CALL(cusparseSbsrmv(
          cusparseHandle, dirA, myCusparseOperation, A.numRows(), A.numCols(),
          A.nnz(), reinterpret_cast<float const*>(&alpha), descrA,
          reinterpret_cast<float const*>(A.values.data()),
          A.graph.row_map.data(), A.graph.entries.data(), A.blockDim(),
          reinterpret_cast<float const*>(x.data()),
          reinterpret_cast<float const*>(&beta),
          reinterpret_cast<float*>(y.data())));
    } else if (std::is_same<value_type, double>::value) {
      KOKKOS_CUSPARSE_SAFE_CALL(cusparseDbsrmv(
          cusparseHandle, dirA, myCusparseOperation, A.numRows(), A.numCols(),
          A.nnz(), reinterpret_cast<double const*>(&alpha), descrA,
          reinterpret_cast<double const*>(A.values.data()),
          A.graph.row_map.data(), A.graph.entries.data(), A.blockDim(),
          reinterpret_cast<double const*>(x.data()),
          reinterpret_cast<double const*>(&beta),
          reinterpret_cast<double*>(y.data())));
    } else if (std::is_same<value_type, Kokkos::complex<float>>::value) {
      KOKKOS_CUSPARSE_SAFE_CALL(cusparseCbsrmv(
          cusparseHandle, dirA, myCusparseOperation, A.numRows(), A.numCols(),
          A.nnz(), reinterpret_cast<cuComplex const*>(&alpha), descrA,
          reinterpret_cast<cuComplex const*>(A.values.data()),
          A.graph.row_map.data(), A.graph.entries.data(), A.blockDim(),
          reinterpret_cast<cuComplex const*>(x.data()),
          reinterpret_cast<cuComplex const*>(&beta),
          reinterpret_cast<cuComplex*>(y.data())));
    } else if (std::is_same<value_type, Kokkos::complex<double>>::value) {
      KOKKOS_CUSPARSE_SAFE_CALL(cusparseZbsrmv(
          cusparseHandle, dirA, myCusparseOperation, A.numRows(), A.numCols(),
          A.nnz(), reinterpret_cast<cuDoubleComplex const*>(&alpha), descrA,
          reinterpret_cast<cuDoubleComplex const*>(A.values.data()),
          A.graph.row_map.data(), A.graph.entries.data(), A.blockDim(),
          reinterpret_cast<cuDoubleComplex const*>(x.data()),
          reinterpret_cast<cuDoubleComplex const*>(&beta),
          reinterpret_cast<cuDoubleComplex*>(y.data())));
    } else {
      throw std::logic_error(
          "Trying to call cusparse[*]bsrmv with a scalar type not "
          "float/double, "
          "nor complex of either!");
    }
  } else {
    throw std::logic_error(
        "With cuSPARSE pre-10.0, offset and entry types must be int. "
        "Something wrong with TPL avail logic.");
  }

  KOKKOS_CUSPARSE_SAFE_CALL(cusparseDestroyMatDescr(descrA));
#endif  // CUDA_VERSION
}

// Reference
// https://docs.nvidia.com/cuda/cusparse/index.html#bsrmm
// Several comments on bsrmm():
// - Only blockDim > 1 is supported
// - Only CUSPARSE_OPERATION_NON_TRANSPOSE is supported
// - Only CUSPARSE_MATRIX_TYPE_GENERAL is supported.
// - Only LayoutLeft for X and Y:
//   for X,Y LayoutLeft we want cuSparse to do
//   C = A * B + C
//   and for X,Y LayoutRight we want cuSparse to do
//   trans(C) = A * trans(B) + trans(C)
//   -> t(t(C)) = t(A * t(B)) + t(t(C))
//   ->       C = t(t(B)) * t(A) + C
//   ->       C = B * t(A) + C
//   This is impossible in cuSparse without explicitly transposing C,
//   so we just do not support LayoutRight in cuSparse TPL now
//
template <
    class AMatrix, class XVector, class YVector,
    std::enable_if_t<std::is_same<Kokkos::LayoutLeft,
                                  typename XVector::array_layout>::value &&
                         std::is_same<Kokkos::LayoutLeft,
                                      typename YVector::array_layout>::value,
                     bool> = true>
void spm_mv_block_impl_cusparse(
    const KokkosKernels::Experimental::Controls& controls, const char mode[],
    typename YVector::non_const_value_type const& alpha, const AMatrix& A,
    const XVector& x, typename YVector::non_const_value_type const& beta,
    const YVector& y) {
  using offset_type = typename AMatrix::non_const_size_type;
  using entry_type  = typename AMatrix::non_const_ordinal_type;
  using value_type  = typename AMatrix::non_const_value_type;

  /* initialize cusparse library */
  cusparseHandle_t cusparseHandle = controls.getCusparseHandle();

  /* Set the operation mode */
  cusparseOperation_t myCusparseOperation;
  switch (toupper(mode[0])) {
    case 'N': myCusparseOperation = CUSPARSE_OPERATION_NON_TRANSPOSE; break;
    default: {
      std::cerr << "Mode " << mode << " invalid for cusparse[*]bsrmv.\n";
      throw std::invalid_argument("Invalid mode");
    }
  }

  int colx = static_cast<int>(x.extent(1));

  // ldx and ldy should be the leading dimension of X,Y respectively
  const int ldx = static_cast<int>(x.extent(0));
  const int ldy = static_cast<int>(y.extent(0));

#if (9000 <= CUDA_VERSION)

  /* create and set the matrix descriptor */
  cusparseMatDescr_t descrA = 0;
  KOKKOS_CUSPARSE_SAFE_CALL(cusparseCreateMatDescr(&descrA));
  KOKKOS_CUSPARSE_SAFE_CALL(
      cusparseSetMatType(descrA, CUSPARSE_MATRIX_TYPE_GENERAL));
  KOKKOS_CUSPARSE_SAFE_CALL(
      cusparseSetMatIndexBase(descrA, CUSPARSE_INDEX_BASE_ZERO));
  cusparseDirection_t dirA = CUSPARSE_DIRECTION_ROW;

  /* perform the actual SpMV operation */
  if ((std::is_same<int, offset_type>::value) &&
      (std::is_same<int, entry_type>::value)) {
    if (std::is_same<value_type, float>::value) {
      KOKKOS_CUSPARSE_SAFE_CALL(cusparseSbsrmm(
          cusparseHandle, dirA, myCusparseOperation,
          CUSPARSE_OPERATION_NON_TRANSPOSE, A.numRows(), colx, A.numCols(),
          A.nnz(), reinterpret_cast<float const*>(&alpha), descrA,
          reinterpret_cast<float const*>(A.values.data()),
          A.graph.row_map.data(), A.graph.entries.data(), A.blockDim(),
          reinterpret_cast<float const*>(x.data()), ldx,
          reinterpret_cast<float const*>(&beta),
          reinterpret_cast<float*>(y.data()), ldy));
    } else if (std::is_same<value_type, double>::value) {
      KOKKOS_CUSPARSE_SAFE_CALL(cusparseDbsrmm(
          cusparseHandle, dirA, myCusparseOperation,
          CUSPARSE_OPERATION_NON_TRANSPOSE, A.numRows(), colx, A.numCols(),
          A.nnz(), reinterpret_cast<double const*>(&alpha), descrA,
          reinterpret_cast<double const*>(A.values.data()),
          A.graph.row_map.data(), A.graph.entries.data(), A.blockDim(),
          reinterpret_cast<double const*>(x.data()), ldx,
          reinterpret_cast<double const*>(&beta),
          reinterpret_cast<double*>(y.data()), ldy));
    } else if (std::is_same<value_type, Kokkos::complex<float>>::value) {
      KOKKOS_CUSPARSE_SAFE_CALL(cusparseCbsrmm(
          cusparseHandle, dirA, myCusparseOperation,
          CUSPARSE_OPERATION_NON_TRANSPOSE, A.numRows(), colx, A.numCols(),
          A.nnz(), reinterpret_cast<cuComplex const*>(&alpha), descrA,
          reinterpret_cast<cuComplex const*>(A.values.data()),
          A.graph.row_map.data(), A.graph.entries.data(), A.blockDim(),
          reinterpret_cast<cuComplex const*>(x.data()), ldx,
          reinterpret_cast<cuComplex const*>(&beta),
          reinterpret_cast<cuComplex*>(y.data()), ldy));
    } else if (std::is_same<value_type, Kokkos::complex<double>>::value) {
      KOKKOS_CUSPARSE_SAFE_CALL(cusparseZbsrmm(
          cusparseHandle, dirA, myCusparseOperation,
          CUSPARSE_OPERATION_NON_TRANSPOSE, A.numRows(), colx, A.numCols(),
          A.nnz(), reinterpret_cast<cuDoubleComplex const*>(&alpha), descrA,
          reinterpret_cast<cuDoubleComplex const*>(A.values.data()),
          A.graph.row_map.data(), A.graph.entries.data(), A.blockDim(),
          reinterpret_cast<cuDoubleComplex const*>(x.data()), ldx,
          reinterpret_cast<cuDoubleComplex const*>(&beta),
          reinterpret_cast<cuDoubleComplex*>(y.data()), ldy));
    } else {
      throw std::logic_error(
          "Trying to call cusparse[*]bsrmm with a scalar type not "
          "float/double, "
          "nor complex of either!");
    }
  } else {
    throw std::logic_error(
        "With cuSPARSE pre-10.0, offset and entry types must be int. "
        "Something wrong with TPL avail logic.");
  }

  KOKKOS_CUSPARSE_SAFE_CALL(cusparseDestroyMatDescr(descrA));
#endif  // CUDA_VERSION
}

#define KOKKOSSPARSE_SPMV_CUSPARSE(SCALAR, ORDINAL, OFFSET, LAYOUT, SPACE,     \
                                   COMPILE_LIBRARY)                            \
  template <>                                                                  \
  struct SPMV_BSRMATRIX<                                                       \
      SCALAR const, ORDINAL const, Kokkos::Device<Kokkos::Cuda, SPACE>,        \
      Kokkos::MemoryTraits<Kokkos::Unmanaged>, OFFSET const, SCALAR const*,    \
      LAYOUT, Kokkos::Device<Kokkos::Cuda, SPACE>,                             \
      Kokkos::MemoryTraits<Kokkos::Unmanaged | Kokkos::RandomAccess>, SCALAR*, \
      LAYOUT, Kokkos::Device<Kokkos::Cuda, SPACE>,                             \
      Kokkos::MemoryTraits<Kokkos::Unmanaged>, true, COMPILE_LIBRARY> {        \
    using device_type       = Kokkos::Device<Kokkos::Cuda, SPACE>;             \
    using memory_trait_type = Kokkos::MemoryTraits<Kokkos::Unmanaged>;         \
    using AMatrix = BsrMatrix<SCALAR const, ORDINAL const, device_type,        \
                              memory_trait_type, OFFSET const>;                \
    using XVector = Kokkos::View<                                              \
        SCALAR const*, LAYOUT, device_type,                                    \
        Kokkos::MemoryTraits<Kokkos::Unmanaged | Kokkos::RandomAccess>>;       \
    using YVector =                                                            \
        Kokkos::View<SCALAR*, LAYOUT, device_type, memory_trait_type>;         \
    using Controls = KokkosKernels::Experimental::Controls;                    \
                                                                               \
    using coefficient_type = typename YVector::non_const_value_type;           \
                                                                               \
    static void spmv_bsrmatrix(const Controls& controls, const char mode[],    \
                               const coefficient_type& alpha,                  \
                               const AMatrix& A, const XVector& x,             \
                               const coefficient_type& beta,                   \
                               const YVector& y) {                             \
      std::string label = "KokkosSparse::spmv[TPL_CUSPARSE,BSRMATRIX" +        \
                          Kokkos::ArithTraits<SCALAR>::name() + "]";           \
      Kokkos::Profiling::pushRegion(label);                                    \
      spmv_block_impl_cusparse(controls, mode, alpha, A, x, beta, y);          \
      Kokkos::Profiling::popRegion();                                          \
    }                                                                          \
  };

#if (9000 <= CUDA_VERSION)
KOKKOSSPARSE_SPMV_CUSPARSE(double, int, int, Kokkos::LayoutLeft,
                           Kokkos::CudaSpace,
                           KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_CUSPARSE(double, int, int, Kokkos::LayoutRight,
                           Kokkos::CudaSpace,
                           KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_CUSPARSE(float, int, int, Kokkos::LayoutLeft,
                           Kokkos::CudaSpace,
                           KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_CUSPARSE(float, int, int, Kokkos::LayoutRight,
                           Kokkos::CudaSpace,
                           KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_CUSPARSE(Kokkos::complex<double>, int, int,
                           Kokkos::LayoutLeft, Kokkos::CudaSpace,
                           KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_CUSPARSE(Kokkos::complex<double>, int, int,
                           Kokkos::LayoutRight, Kokkos::CudaSpace,
                           KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_CUSPARSE(Kokkos::complex<float>, int, int, Kokkos::LayoutLeft,
                           Kokkos::CudaSpace,
                           KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_CUSPARSE(Kokkos::complex<float>, int, int,
                           Kokkos::LayoutRight, Kokkos::CudaSpace,
                           KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_CUSPARSE(double, int, int, Kokkos::LayoutLeft,
                           Kokkos::CudaUVMSpace,
                           KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_CUSPARSE(double, int, int, Kokkos::LayoutRight,
                           Kokkos::CudaUVMSpace,
                           KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_CUSPARSE(float, int, int, Kokkos::LayoutLeft,
                           Kokkos::CudaUVMSpace,
                           KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_CUSPARSE(float, int, int, Kokkos::LayoutRight,
                           Kokkos::CudaUVMSpace,
                           KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_CUSPARSE(Kokkos::complex<double>, int, int,
                           Kokkos::LayoutLeft, Kokkos::CudaUVMSpace,
                           KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_CUSPARSE(Kokkos::complex<double>, int, int,
                           Kokkos::LayoutRight, Kokkos::CudaUVMSpace,
                           KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_CUSPARSE(Kokkos::complex<float>, int, int, Kokkos::LayoutLeft,
                           Kokkos::CudaUVMSpace,
                           KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
KOKKOSSPARSE_SPMV_CUSPARSE(Kokkos::complex<float>, int, int,
                           Kokkos::LayoutRight, Kokkos::CudaUVMSpace,
                           KOKKOSKERNELS_IMPL_COMPILE_LIBRARY)
#endif  // 9000 <= CUDA_VERSION

#undef KOKKOSSPARSE_SPMV_CUSPARSE

// cuSparse TPL does not support LayoutRight for this operation
// only specialize for LayoutLeft
#define KOKKOSSPARSE_SPMV_MV_CUSPARSE(SCALAR, ORDINAL, OFFSET, SPACE,          \
                                      ETI_AVAIL)                               \
  template <>                                                                  \
  struct SPMV_MV_BSRMATRIX<                                                    \
      SCALAR const, ORDINAL const, Kokkos::Device<Kokkos::Cuda, SPACE>,        \
      Kokkos::MemoryTraits<Kokkos::Unmanaged>, OFFSET const, SCALAR const**,   \
      Kokkos::LayoutLeft, Kokkos::Device<Kokkos::Cuda, SPACE>,                 \
      Kokkos::MemoryTraits<Kokkos::Unmanaged | Kokkos::RandomAccess>,          \
      SCALAR**, Kokkos::LayoutLeft, Kokkos::Device<Kokkos::Cuda, SPACE>,       \
      Kokkos::MemoryTraits<Kokkos::Unmanaged>, false, true, ETI_AVAIL> {       \
    using device_type       = Kokkos::Device<Kokkos::Cuda, SPACE>;             \
    using memory_trait_type = Kokkos::MemoryTraits<Kokkos::Unmanaged>;         \
    using AMatrix = BsrMatrix<SCALAR const, ORDINAL const, device_type,        \
                              memory_trait_type, OFFSET const>;                \
    using XVector = Kokkos::View<                                              \
        SCALAR const**, Kokkos::LayoutLeft, device_type,                       \
        Kokkos::MemoryTraits<Kokkos::Unmanaged | Kokkos::RandomAccess>>;       \
    using YVector  = Kokkos::View<SCALAR**, Kokkos::LayoutLeft, device_type,   \
                                 memory_trait_type>;                          \
    using Controls = KokkosKernels::Experimental::Controls;                    \
                                                                               \
    using coefficient_type = typename YVector::non_const_value_type;           \
                                                                               \
    static void spmv_mv_bsrmatrix(const Controls& controls, const char mode[], \
                                  const coefficient_type& alpha,               \
                                  const AMatrix& A, const XVector& x,          \
                                  const coefficient_type& beta,                \
                                  const YVector& y) {                          \
      std::string label = "KokkosSparse::spmv[TPL_CUSPARSE,BSRMATRIX" +        \
                          Kokkos::ArithTraits<SCALAR>::name() + "]";           \
      Kokkos::Profiling::pushRegion(label);                                    \
      spm_mv_block_impl_cusparse(controls, mode, alpha, A, x, beta, y);        \
      Kokkos::Profiling::popRegion();                                          \
    }                                                                          \
  };

#if (9000 <= CUDA_VERSION)
KOKKOSSPARSE_SPMV_MV_CUSPARSE(double, int, int, Kokkos::CudaSpace, true)
KOKKOSSPARSE_SPMV_MV_CUSPARSE(double, int, int, Kokkos::CudaSpace, false)
KOKKOSSPARSE_SPMV_MV_CUSPARSE(float, int, int, Kokkos::CudaSpace, true)
KOKKOSSPARSE_SPMV_MV_CUSPARSE(float, int, int, Kokkos::CudaSpace, false)
KOKKOSSPARSE_SPMV_MV_CUSPARSE(Kokkos::complex<double>, int, int,
                              Kokkos::CudaSpace, true)
KOKKOSSPARSE_SPMV_MV_CUSPARSE(Kokkos::complex<double>, int, int,
                              Kokkos::CudaSpace, false)
KOKKOSSPARSE_SPMV_MV_CUSPARSE(Kokkos::complex<float>, int, int,
                              Kokkos::CudaSpace, true)
KOKKOSSPARSE_SPMV_MV_CUSPARSE(Kokkos::complex<float>, int, int,
                              Kokkos::CudaSpace, false)
KOKKOSSPARSE_SPMV_MV_CUSPARSE(double, int, int, Kokkos::CudaUVMSpace, true)
KOKKOSSPARSE_SPMV_MV_CUSPARSE(double, int, int, Kokkos::CudaUVMSpace, false)
KOKKOSSPARSE_SPMV_MV_CUSPARSE(float, int, int, Kokkos::CudaUVMSpace, true)
KOKKOSSPARSE_SPMV_MV_CUSPARSE(float, int, int, Kokkos::CudaUVMSpace, false)
KOKKOSSPARSE_SPMV_MV_CUSPARSE(Kokkos::complex<double>, int, int,
                              Kokkos::CudaUVMSpace, true)
KOKKOSSPARSE_SPMV_MV_CUSPARSE(Kokkos::complex<double>, int, int,
                              Kokkos::CudaUVMSpace, false)
KOKKOSSPARSE_SPMV_MV_CUSPARSE(Kokkos::complex<float>, int, int,
                              Kokkos::CudaUVMSpace, true)
KOKKOSSPARSE_SPMV_MV_CUSPARSE(Kokkos::complex<float>, int, int,
                              Kokkos::CudaUVMSpace, false)

#endif  // 9000 <= CUDA_VERSION

#undef KOKKOSSPARSE_SPMV_MV_CUSPARSE

}  // namespace Impl
}  // namespace Experimental
}  // namespace KokkosSparse

#endif  // KOKKOSKERNELS_ENABLE_TPL_CUSPARSE

#endif  // KOKKOSSPARSE_SPMV_BSRMATRIX_TPL_SPEC_DECL_HPP
